/*===- TableGen'erated file -------------------------------------*- C++ -*-===*\
|*                                                                            *|
|* Operation Interface Definitions                                            *|
|*                                                                            *|
|* Automatically generated file, do not edit!                                 *|
|*                                                                            *|
\*===----------------------------------------------------------------------===*/

void OpAsmOpInterface::getAsmResultNames(OpAsmSetValueNameFn setNameFn) {
      return getImpl()->getAsmResultNames(getOperation(), setNameFn);
  }
