/*===- TableGen'erated file -------------------------------------*- C++ -*-===*\
|*                                                                            *|
|* Dialect Declarations                                                       *|
|*                                                                            *|
|* Automatically generated file, do not edit!                                 *|
|*                                                                            *|
\*===----------------------------------------------------------------------===*/


class SCFDialect : public ::mlir::Dialect {
public:
  explicit SCFDialect(::mlir::MLIRContext *context);
  static ::llvm::StringRef getDialectNamespace() { return "scf"; }
};
